from abc import ABC, abstractmethod
from contextlib import contextmanager
from contextvars import ContextVar
from functools import singledispatch
from typing import (
    Any,
    Callable,
    ClassVar,
    Literal,
    Optional,
    Protocol,
    Union,
    overload,
)
import uuid
from .rendering import Rendering
from .listdiff import Reorder, diff as listdiff
from .patch import InvalidatePatch, Patch


class VdomContext(Protocol):
    def _patch(self, patch: Patch) -> None:
        ...

    def _register_event(self, k: str, handler: Callable):
        ...

    def _unregister_event(self, k: str):
        ...


vdom_context: ContextVar[VdomContext] = ContextVar("vdom_context")


@contextmanager
def set_vdom_context(r: VdomContext):
    t = vdom_context.set(r)
    try:
        yield r
    finally:
        vdom_context.reset(t)


def patch(patch: Patch):
    return vdom_context.get()._patch(patch)


class Vdom(ABC):
    spec_type: ClassVar[type]
    key: str

    @abstractmethod
    def __str__(self) -> str:
        ...

    @abstractmethod
    def render(self) -> Rendering:
        ...

    @abstractmethod
    def dispose(self):
        ...

    @abstractmethod
    def reconcile(self, new_spec: "NormSpec") -> "Vdom":
        ...


class NormSpec(ABC):
    key: int

    @abstractmethod
    def __str__(self):
        ...

    @abstractmethod
    def create(self) -> Vdom:
        ...


@overload
def create(s: NormSpec) -> Vdom:
    ...


@overload
def create(s: list[NormSpec]) -> list[Vdom]:
    ...


def create(s):
    if isinstance(s, list):
        return list(map(create, s))
    elif isinstance(s, NormSpec):
        return s.create()
    else:
        raise TypeError(f"unrecognised spec {s}")


@overload
def render(s: Vdom) -> Rendering:
    ...


@overload
def render(s: list[Vdom]) -> list[Rendering]:
    ...


def render(s: Union[Vdom, list[Vdom]]):
    if isinstance(s, list):
        return list(map(render, s))
    elif isinstance(s, Vdom):
        return s.render()
    else:
        raise TypeError(f"unrecognised spec {s}")


def reconcile_lists(
    old: list[Vdom], new: list[NormSpec]
) -> tuple[list[Vdom], Reorder[Vdom]]:
    r: Reorder = listdiff([x.key for x in old], [x.key for x in new])
    for ri in r.deletions:
        old[ri].dispose()
    new_vdom: list[Any] = [None] * len(new)
    for i, j in r.moves:
        assert new_vdom[j] is None
        new_vdom[j] = reconcile(old[i], new[j])
    for j in r.creations:
        assert new_vdom[j] is None
        new_vdom[j] = new[j].create()
    assert all(x is not None for x in new_vdom)
    r = r.map_inserts(lambda j, _: new_vdom[j])
    return new_vdom, r


def reconcile(old: Vdom, new: NormSpec) -> Vdom:
    if isinstance(new, old.spec_type):
        return old.reconcile(new)  # type: ignore
    old.dispose()
    # [todo] add better patch... this shouldn't really happen because we match on keys.
    patch(InvalidatePatch())
    return new.create()


def dispose(v: Union[Vdom, list["Vdom"]]):
    if isinstance(v, list):
        for vv in v:
            vv.dispose()
    else:
        v.dispose()


Id = str

UUID = uuid.uuid4().hex[:4]
ID_COUNTER = 100


def fresh_id() -> Id:
    global ID_COUNTER
    ID_COUNTER += 1
    return Id(f"{UUID}-{ID_COUNTER}")


Html = Optional[Union[str, list["Html"], Literal[False], NormSpec]]


@singledispatch
def to_html(x: Any):
    raise NotImplementedError(f"Unknown {type(x).__name__}")


def normalise_html(c: Html) -> list[NormSpec]:
    def norm_list(cs) -> Any:
        for c in cs:
            if isinstance(c, str):
                yield to_html(c)
            elif isinstance(c, (list, tuple)):
                yield from norm_list(c)
            elif c is None or c is False or not c:
                continue
            elif isinstance(c, NormSpec):
                yield c
            elif hasattr(c, "__html__"):
                yield from normalise_html(c.__html__())
            else:
                yield to_html(c)

    return list(norm_list([c]))



from uxu.listdiff import diff, Reorder
from hypothesis import given, assume
import hypothesis.strategies as hs
from itertools import product


def texts():
    return hs.text("abcdefg")


@hs.composite
def expanders(draw):
    l1 = list(draw(texts()))
    l2 = list(draw(texts()))
    indices = draw(hs.sets(hs.integers(min_value=0, max_value=len(l1))))
    expander = {i: draw(hs.integers(min_value=0, max_value=10)) for i in indices}
    return l1, l2, expander


def flatten(l):
    def core(x):
        if isinstance(x, list):
            for y in x:
                yield from core(y)
        else:
            yield x

    return list(core(l))


def expander_prop(xs):
    l1, l2, expander = xs
    δ = diff(l1, l2)
    δe = δ.expand_elements(expander)
    l2i = δ.apply(list(range(len(l1))))
    assert len(l2) == len(l2i)

    l1e = flatten(
        [
            [f"{x}-{k}" for k in range(expander[i])] if i in expander else x
            for i, x in enumerate(l1)
        ]
    )
    l2e = flatten(
        [
            [f"{l2[j]}-{k}" for k in range(expander[x])]
            if (isinstance(x, int) and x in expander)
            else l2[j]
            for j, x in enumerate(l2i)
        ]
    )

    l2ex = δe.apply(l1e)

    assert len(l2ex) == len(l2e)
    assert all(x == y for x, y in zip(l2ex, l2e))


def validate_prop(l1, l2):
    r = diff(l1, l2)
    r.validate()


def recover_prop(l1, l2):
    r = diff(l1, l2)
    l3 = r.apply(l1)
    assert len(l2) == len(l3)
    assert all(x is y for x, y in zip(l2, l3))


def invert_prop(l1, l2):
    r = diff(l1, l2).compose(diff(l2, l1)).apply(l1)
    assert len(l1) == len(r)
    assert all(x is y for x, y in zip(l1, r))


def compose_prop(l1, l2, l3):
    δ12 = diff(l1, l2)
    δ23 = diff(l2, l3)

    δ13x = δ12.compose(δ23)
    δ13x.validate()

    l3x = δ13x.apply(l1)
    assert len(l3) == len(l3x)
    assert all(x is y for x, y in zip(l3, l3x))


def concat_prop(l1, l2, l3, l4):
    δ1_2 = diff(l1, l2)
    δ3_4 = diff(l3, l4)
    δ13_24 = δ1_2.concat(δ3_4)
    δ13_24.validate()
    r1 = δ13_24.apply(l1 + l3)
    r2 = l2 + l4
    assert len(r1) == len(r2)
    assert all(x is y for x, y in zip(r1, r2))


examples = ["", "a", "aa", "aaa", "ab", "ba", "baa", "bba", "aab"]
examples = [list(x) for x in examples]


def test_recover_explicit():
    for x, y in product(examples, examples):
        recover_prop(x, y)


def test_validate_explicit():
    for x, y in product(examples, examples):
        validate_prop(x, y)


def test_prop_compose():
    for x, y, z in product(examples, examples, examples):
        compose_prop(x, y, z)


def test_prop_concat():
    for x, y, z, w in product(examples, examples, examples, examples):
        concat_prop(x, y, z, w)


def test_expander1():
    x = (["a"], ["a"], {0: 2})
    expander_prop(x)


@given(texts(), texts())
def test_recover_chars(x1: str, x2: str):
    recover_prop(list(x1), list(x2))


@given(texts(), texts(), texts())
def test_compose_chars(x1: str, x2: str, x3: str):
    compose_prop(list(x1), list(x2), list(x3))


@given(texts(), texts(), texts(), texts())
def test_concat_chars(x1: str, x2: str, x3: str, x4: str):
    concat_prop(list(x1), list(x2), list(x3), list(x4))


@given(texts(), texts())
def test_validate_chars(x1: str, x2: str):
    validate_prop(list(x1), list(x2))


@given(expanders())
def test_expander_chars(x):
    expander_prop(x)


if __name__ == "__main__":
    for x in examples:
        for y in examples:
            r = diff(x, y)
            r.apply(x)



from dataclasses import dataclass
from typing import ClassVar

from .rendering import RenderedText, Rendering
from .patch import InvalidatePatch
from .vdom import Id, NormSpec, Vdom, patch, fresh_id, to_html


@dataclass
class TextNodeSpec(NormSpec):
    value: str

    def create(self):
        return TextNode(key=self.key, id=fresh_id(), value=self.value)

    @property
    def key(self):
        return hash(("text", self.value))

    def __str__(self):
        return f"TextNodeSpec({self.value})"


@dataclass
class TextNode(Vdom):
    spec_type: ClassVar = TextNodeSpec
    key: int
    id: Id
    value: str

    def __str__(self) -> str:
        return self.value

    def dispose(self):
        return

    def render(self) -> Rendering:
        return RenderedText(self.value, self.id)

    def reconcile(self, new_spec: TextNodeSpec) -> "TextNode":
        assert isinstance(new_spec, TextNodeSpec)
        # [todo] patch
        patch(InvalidatePatch())
        self.dispose()
        return new_spec.create()


@to_html.register(str)
def to_html_str(value: str):
    return TextNodeSpec(value)



from dataclasses import dataclass, field, replace
from typing import Any, Callable, Union

""" A rendering is the thing that we actually send over the wire to Javascript. """


@dataclass
class EventHandler:
    handler_id: Any

    def __eq__(self, other):
        if not isinstance(other, EventHandler):
            return False
        return self.handler_id == other.handler_id

    def __todict__(self):
        return {"__handler__": self.handler_id}

    @classmethod
    def __ofdict__(cls, d: dict):
        return EventHandler(d["__handler__"])


RenderedAttrVal = Union[EventHandler, str, dict]


@dataclass
class RenderedText:
    value: str
    id: Any
    kind: str = field(default="text")


@dataclass
class RenderedElement:
    id: Any
    tag: str
    attrs: dict[str, RenderedAttrVal]
    children: list["Rendering"]
    kind: str = field(default="element")


@dataclass
class RenderedFragment:
    id: Any
    children: list["Rendering"]
    kind: str = field(default="fragment")


@dataclass
class RenderedWidget:
    """This is used to hook into JavaScript code."""

    id: Any
    name: str
    props: Any
    kind: str = field(default="widget")


Rendering = Union[RenderedElement, RenderedText, RenderedFragment]


def iter_event_handlers(x: "Rendering"):
    if isinstance(x, RenderedElement):
        for name, v in x.attrs.items():
            if isinstance(v, EventHandler):
                yield name, v
    children = getattr(x, "children", [])
    for child in children:
        yield from iter_event_handlers(child)


def map_event_handlers(modify: Callable[[EventHandler], EventHandler]):
    def rec(x: Rendering) -> Rendering:
        if isinstance(x, RenderedElement):
            attrs = {
                k: modify(v) if isinstance(v, EventHandler) else v
                for k, v in x.attrs.items()
            }
            children = list(map(rec, x.children))
            return replace(x, attrs=attrs, children=children)
        elif isinstance(x, RenderedText):
            return x
        elif isinstance(x, RenderedFragment):
            return replace(x, children=list(map(rec, x.children)))
        else:
            raise TypeError()

    return rec



import asyncio
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Protocol, ClassVar

from .rendering import RenderedText, iter_event_handlers

from .manager import EventArgs
from .patch import Patch, InvalidatePatch
from .vdom import NormSpec, Vdom, Rendering, fresh_id, vdom_context, patch


class ReactorLike(Protocol):
    async def wait_patches(self) -> list[Patch]:
        ...

    async def initialize(self):
        ...

    async def render(self) -> Rendering:
        ...

    def handle_event(self, args: EventArgs):
        ...

    def dispose(self):
        ...


@dataclass
class ReactorVdom:
    id: Any
    rlike: ReactorLike

    def render(self):
        r = getattr(
            self, "rendered", RenderedText(value="loading", id=f"{self.id}-loading")
        )
        return r

    def dispose(self):
        self.main_task.cancel()
        self.rlike.dispose()

    @classmethod
    def create(cls, rlike):
        x = cls(
            id=fresh_id(),
            rlike=rlike,
        )
        x._start()
        return x

    async def refresh(self):
        rendered = await self.rlike.render()
        old_handler_ids = getattr(self, "handler_ids", set())
        handler_ids = set()
        for name, eh in iter_event_handlers(rendered):
            handler_ids.add(eh.handler_id)
            old_handler_ids.discard(eh.handler_id)
            vdom_context.get()._register_event(
                eh.handler_id, partial(self.handle_event, eh.handler_id, name)
            )
        for ohid in old_handler_ids:
            vdom_context.get()._unregister_event(ohid)
        self.rendered = rendered
        self.handler_ids = handler_ids
        # [todo] do fancy patch forwarding
        patch(InvalidatePatch())

    def handle_event(self, handler_id, name, params):
        self.rlike.handle_event(
            EventArgs(handler_id=handler_id, name=name, params=params)
        )

    def _start(self):
        self.main_task = asyncio.create_task(self._patcher_loop())

    async def _patcher_loop(self):
        await self.rlike.initialize()
        await self.refresh()
        while True:
            patches = await self.rlike.wait_patches()
            if len(patches) > 0:
                # [todo] forward the patches here.
                await self.refresh()


class ThunkSpec(NormSpec):
    def __init__(self, thunk: Callable[[], Vdom], key=None):
        self.name = thunk.__name__
        self.key = hash(f"ThunkSpec {self.name} {key}")
        self.thunk = thunk

    def create(self):
        return ThunkVdom(wrapped=self.thunk(), name=self.name, key=self.key)

    def __str__(self):
        return f"ThunkSpec {self.name}"


@dataclass
class ThunkVdom(Vdom):
    spec_type: ClassVar = ThunkSpec
    wrapped: Vdom
    name: str
    key: Any

    def reconcile(self, spec: ThunkSpec):
        if spec.name == self.name:
            return self
        else:
            self.dispose()
            return spec.create()

    def __str__(self):
        return f"ThunkVdom {self.name}"

    def dispose(self):
        return self.wrapped.dispose()

    def render(self):
        return self.wrapped.render()



import asyncio
from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    ClassVar,
    Coroutine,
    Generic,
    Optional,
    ParamSpec,
    Protocol,
    TypeVar,
    Union,
    overload,
)
from .rendering import RenderedFragment, Rendering
from .patch import ModifyChildrenPatch
from .vdom import (
    Html,
    NormSpec,
    Vdom,
    VdomContext,
    create,
    dispose,
    fresh_id,
    normalise_html,
    patch,
    vdom_context,
    reconcile_lists,
)
import logging

logger = logging.getLogger("uxu")


S = TypeVar("S")

SetterFn = Callable[[Union[S, Callable[[S], S]]], None]


class NoNeedToRerender(Exception):
    pass


class AbstractHook(Protocol):
    def reconcile(self, new_hook) -> "AbstractHook":
        ...

    def dispose(self) -> None:
        ...

    def initialize(self) -> None:
        ...


class StateHook(Generic[S]):
    state: S
    fiber: Optional["Fiber"]

    def __init__(self, init: S, fiber: "Fiber"):
        self.state = init
        self.fiber = fiber

    def __str__(self):
        return f"<{type(self).__name__} {type(self.state)}>"

    def reconcile(self, new_hook: "StateHook") -> "StateHook":
        assert type(self) == type(new_hook)
        return self

    def dispose(self):
        self.fiber = None

    @property
    def current(self) -> S:
        """Alias for `self.state`"""
        return self.state

    def invalidate(self):
        if self.fiber is not None:
            self.fiber.invalidate()

    def modify(self, fn: Callable[[S], S]) -> None:
        old_state = self.state
        self.state = fn(old_state)
        # [todo] add equality check
        self.invalidate()
        logger.debug(f"{str(self.fiber)}: {old_state} -> {self.state}")
        return

    def initialize(self) -> None:
        return

    def set(self, item: S) -> None:
        return self.modify(lambda _: item)


class EffectHook:
    def __init__(self, callback, deps):
        self.task = None
        self.callback = callback
        self.deps = deps

    def __str__(self):
        return f"<{type(self).__name__}>"

    def dispose(self):
        if self.task is not None:
            self.task.cancel()
            self.task = None

    def initialize(self):
        self.evaluate()

    def evaluate(self):
        callback = self.callback
        assert callable(callback)
        self.dispose()
        if asyncio.iscoroutinefunction(callback):
            self.task = asyncio.create_task(callback())
        else:
            callback()

    def reconcile(self, new_hook: "EffectHook"):
        assert type(new_hook) == type(self)

        def update():
            self.deps = new_hook.deps
            self.callback = new_hook.callback
            self.evaluate()

        if self.deps is None and new_hook.deps is None:
            update()
        elif len(self.deps) != len(new_hook.deps):
            update()
        else:
            for old_dep, new_dep in zip(self.deps, new_hook.deps):
                if old_dep != new_dep:
                    logger.debug(f"Dep changed {old_dep} -> {new_dep}")
                    update()
                    break
        return self


P = ParamSpec("P")


class Component(Protocol[P]):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Html:
        ...


@dataclass
class FiberSpec(Generic[P], NormSpec):
    component: Component[P]
    props_args: list
    props_kwargs: dict
    key: Optional[str] = field(default=None)

    @property
    def name(self):
        return getattr(self.component, "__name__", "unknown")

    def create(self):
        return Fiber.create(self)

    def __str__(self):
        return self.name


H = TypeVar("H", bound="AbstractHook")


class Fiber(Vdom):
    """Like React fibers."""

    spec_type: ClassVar = FiberSpec

    component: "Component"
    props_args: list
    props_kwargs: dict

    hooks: list[AbstractHook]
    hook_idx: int
    rendered: list[Vdom]
    invalidated_event: asyncio.Event
    update_loop_task: asyncio.Task

    @property
    def name(self) -> str:
        return getattr(self.component, "__name__")

    def __str__(self) -> str:
        return f"<{self.name} {self.id}>"

    def __init__(self, spec: "FiberSpec"):
        self.id = fresh_id()
        self.key = spec.key  # type: ignore
        self.invalidated_event = asyncio.Event()
        component = spec.component
        if not hasattr(component, "__name__"):
            logger.warning(f"Please name component {component}.")
        self.component = component
        self.props_args = spec.props_args
        self.props_kwargs = spec.props_kwargs
        assert not hasattr(self, "rendered") and not hasattr(
            self, "hooks"
        ), "already created"
        self.hooks = []
        self.hook_idx = 0
        t = fiber_context.set(self)
        s = self.component(*self.props_args, **self.props_kwargs)
        self.rendered = create(normalise_html(s))
        fiber_context.reset(t)
        self.update_loop_task = asyncio.create_task(self.update_loop())

    async def update_loop(self):
        while True:
            await self.invalidated_event.wait()
            logger.debug(f"{str(self)} rerendering.")
            self.invalidated_event.clear()
            try:
                self.reconcile_core()
            except Exception as e:
                logger.exception("failure in update loop")

    def dispose(self):
        # [todo] can I just use GC?
        assert hasattr(self, "hooks")
        assert hasattr(self, "rendered"), "fiber is not rendered"
        dispose(self.rendered)
        for hook in reversed(self.hooks):
            hook.dispose()
        self.update_loop_task.cancel()

    def invalidate(self):
        """Called when a hook's callback is invoked, means that a re-render must occur."""
        logger.debug(f"{str(self)} invalidated.")
        self.invalidated_event.set()

    def reconcile_hook(self, hook: H) -> H:
        if self.hook_idx >= len(self.hooks):
            # initialisation case
            self.hooks.append(hook)
            self.hook_idx += 1
            hook.initialize()
            return hook

        old_hook = self.hooks[self.hook_idx]
        if type(old_hook) != type(hook):
            logger.error(
                f"{self} {self.hook_idx}th hook changed type from {str(old_hook)} to {str(hook)}"
            )
            old_hook.dispose()
            self.hooks[self.hook_idx] = hook
            hook.initialize()
            self.hook_idx += 1
            return hook
        else:
            self.hook_idx += 1
            return old_hook.reconcile(hook)  # type: ignore

    def reconcile_core(self):
        self.invalidated_event.clear()
        t = fiber_context.set(self)
        self.hook_idx = 0
        try:
            spec = self.component(*self.props_args, **self.props_kwargs)
        except NoNeedToRerender:
            assert hasattr(self, "rendered")
            logger.debug(f"{str(self)} Skipping re-render")
            return
        except Exception as e:
            logger.exception(f"{self} error while rendering")
            # [todo] inject a message into DOM here, or set the border colour.
            return
        finally:
            fiber_context.reset(t)
        spec = normalise_html(spec)
        children, reorder = reconcile_lists(self.rendered, spec)
        self.rendered = children
        l = self.hook_idx + 1
        old_hooks = self.hooks[l:]
        self.hooks = self.hooks[:l]
        for hook in reversed(old_hooks):
            hook.dispose()
        patch(
            ModifyChildrenPatch(
                element_id=self.id,
                remove_these=reorder.remove_these,
                then_insert_these=reorder.then_insert_these,
                children_length_start=reorder.l1_len,
            )
        )
        return

    def reconcile(self, new_spec: "FiberSpec") -> "Fiber":
        assert isinstance(new_spec, FiberSpec)
        assert hasattr(self, "hooks") and hasattr(self, "rendered"), "not created"
        # if the identity of the component function has changed that
        # means we should rerender.
        if new_spec.name != self.name or self.component is not new_spec.component:
            self.dispose()
            new_fiber = Fiber(new_spec)
            return new_fiber
        # [todo] check whether the props have changed here
        if (
            self.props_args == new_spec.props_args
            and self.props_kwargs == new_spec.props_kwargs
            and not self.invalidated_event.is_set()
        ):
            logger.debug(f"{str(self)} has unchanged props. Skipping re-render.")
            return self
        self.props_args = new_spec.props_args
        self.props_kwargs = new_spec.props_kwargs
        self.reconcile_core()
        return self

    def render(self) -> Rendering:
        return RenderedFragment(
            id=self.id, children=[x.render() for x in self.rendered]
        )

    @classmethod
    def create(cls, spec: FiberSpec):
        return cls(spec)


fiber_context: ContextVar[Fiber] = ContextVar("fiber_context")


class StateVar(Protocol[S]):
    @property
    def current(self) -> S:
        ...

    def set(self, item: S):
        ...

    def modify(self, fn: Callable[[S], S]):
        ...

    def invalidate(self):
        ...


def useState(init: S) -> StateVar[S]:
    ctx = fiber_context.get(None)
    if ctx is None:
        raise RuntimeError("useState can only be used inside a component function.")
    return ctx.reconcile_hook(StateHook(init, ctx))


def useEffect(callback: Callable[[], Coroutine[Any, Any, None]], deps=None):
    ctx = fiber_context.get()
    ctx.reconcile_hook(EffectHook(callback, deps))



from collections import Counter, defaultdict
from dataclasses import dataclass, replace
from enum import Enum
from functools import partial
from typing import Any, Callable, Generic, Iterable, Optional, TypeVar, Union
from difflib import SequenceMatcher

from miniscutil import map_keys, Sum

A = TypeVar("A")
B = TypeVar("B")
R = TypeVar("R")


@dataclass
class Reorder(Generic[A]):
    """Represents a patch of removing and inserting items between a pair of `list[A]`s.

    You should not instantiate this yourself, instead use the `diff` function.
    """

    l1_len: int
    l2_len: int
    remove_these: dict[int, Optional[str]]
    then_insert_these: dict[int, Sum[str, A]]

    def apply(self, l1: list[A]) -> list[A]:
        if len(l1) < self.l1_len:
            raise ValueError(f"list length ({len(l1)}) must be at least {self.l1_len}")
        l2 = []
        removed = {}
        for i in range(len(l1)):
            if i in self.remove_these:
                k = self.remove_these[i]
                if k is not None:
                    removed[k] = l1[i]
            else:
                l2.append(l1[i])
        for j in sorted(self.then_insert_these.keys()):
            o = self.then_insert_these[j].match(lambda k: removed[k], lambda t: t)
            l2.insert(j, o)
        return l2

    def map_inserts(self, fn: Callable[[int, A], B]) -> "Reorder[B]":
        x = {j: v.mapr(partial(fn, j)) for j, v in self.then_insert_these.items()}
        return replace(self, then_insert_these=x)  # type: ignore

    def validate(self):
        assert all(isinstance(i, int) for i in self.remove_these.keys())
        assert all(isinstance(j, int) for j in self.then_insert_these.keys())
        assert all(
            j < self.l2_len for j in self.then_insert_these.keys()
        ), "then_insert_these must be in range"
        assert all(
            i < self.l1_len for i in self.remove_these.keys()
        ), "remove_these must be in range"
        rm_keys = [v for v in self.remove_these.values() if v is not None]
        assert len(set(rm_keys)) == len(rm_keys), "remove_these keys must be unique"
        assert all(
            isinstance(v, Sum) for v in self.then_insert_these.values()
        ), "then_insert_these are not Sum types"
        in_keys = [v.value for v in self.then_insert_these.values() if v.is_left]
        assert len(set(in_keys)) == len(
            in_keys
        ), "then_insert_these keys must be unique"
        assert set(in_keys) == set(
            rm_keys
        ), "then_insert_these keys must match remove_these keys"

    def compose(self, after: "Reorder[A]", extend=False) -> "Reorder[A]":
        """Create a `Reorder` that is equivalent to applying `self` followed by `after`."""

        if self.l2_len != after.l1_len:
            n1 = self.l2_len
            n2 = after.l1_len
            if extend:
                if n1 < n2:
                    return self.extend_right(n2 - n1).compose(after, extend=False)
                else:
                    return self.compose(after.extend_right(n1 - n2), extend=False)
            raise ValueError(f"Intermediate lengths do not match")

        δ1: Reorder[Sum[int, A]] = self.map_inserts(lambda j, a: Sum.inr(a))
        δ2: Reorder[Sum[int, A]] = after.map_inserts(lambda k, a: Sum.inr(a))
        l1 = [Sum.inl(i) for i in range(self.l1_len)]
        l2: list[Sum[int, A]] = δ1.apply(l1)
        remove_these = {
            i: (f"1-{v}" if v is not None else None)
            for i, v in self.remove_these.items()
        }
        for j, v in after.remove_these.items():
            s = l2[j]
            if s.is_left:
                i = s.value
                assert isinstance(i, int)
                remove_these[i] = f"2-{v}" if v is not None else None
        l2 = [Sum.inl(j) for j in range(self.l2_len)]
        l3: list[Sum[int, A]] = δ2.apply(l2)
        insert_these = {
            k: v.map(left=lambda v: f"2-{v}", right=lambda a: a)
            for k, v in after.then_insert_these.items()
        }
        for j, v in δ1.then_insert_these.items():
            try:
                k = l3.index(Sum.inl(j))
            except ValueError:
                continue
            if v.is_left:
                s = v.value
                assert isinstance(s, str), f"expected string but got {s} : {type(s)}"
                if k in insert_these:
                    ss = insert_these[k]
                    assert ss.is_left
                else:
                    insert_these[k] = Sum.inl(f"1-{s}")
            else:
                ja: Sum[int, A] = v.value
                assert not ja.is_left
                a: A = ja.value
                insert_these[k] = Sum.inr(a)

        return Reorder(
            l1_len=self.l1_len,
            l2_len=after.l2_len,
            remove_these=remove_these,
            then_insert_these=insert_these,
        )

    def extend_left(self, length: int) -> "Reorder[A]":
        return Reorder(
            length + self.l1_len,
            length + self.l2_len,
            remove_these=map_keys(lambda i: i + length, self.remove_these),
            then_insert_these=map_keys(lambda j: j + length, self.then_insert_these),
        )

    def extend_right(self, length: int) -> "Reorder[A]":
        return Reorder(
            length + self.l1_len,
            length + self.l2_len,
            remove_these=self.remove_these,
            then_insert_these=self.then_insert_these,
        )

    def concat(self, other: "Reorder[A]") -> "Reorder[A]":
        """Horizontally concatenate `self` and `other`."""
        δ1 = self._map_keys(lambda s: f"a-{s}")
        δ2 = other._map_keys(lambda s: f"b-{s}")
        r2 = map_keys(lambda i: δ1.l1_len + i, δ2.remove_these)
        i2 = map_keys(lambda j: δ1.l2_len + j, δ2.then_insert_these)
        r = {**δ1.remove_these, **r2}
        i = {**δ1.then_insert_these, **i2}

        return Reorder(
            self.l1_len + other.l1_len,
            self.l2_len + other.l2_len,
            remove_these=r,
            then_insert_these=i,
        )

    def expand_elements(self, i_to_len: dict[int, int]) -> "Reorder[A]":
        """Return the reordering when the given l1 indices are expanded at the given index."""
        bloat = sum(i_to_len.values())
        nr = 0
        ni = 0
        r = {}
        ins = {}
        for i in range(self.l1_len):
            if i in i_to_len:
                l = i_to_len[i]
                if i in self.remove_these:
                    v = self.remove_these[i]
                    if v is not None:
                        for k in range(l):
                            r[i + nr + k] = f"{k}-{v}"
                    else:
                        for k in range(l):
                            r[i + nr + k] = None
                nr += l - 1
            elif i in self.remove_these:
                r[i + nr] = self.remove_these[i]
        l2 = self.map_inserts(lambda j, a: Sum.inr(a)).apply(list(map(Sum.inl, range(self.l1_len))))  # type: ignore
        for j, ia in enumerate(l2):
            if ia.is_left:
                i: int = ia.value
                if i in i_to_len:
                    l = i_to_len[i]
                    if j in self.then_insert_these:
                        ka = self.then_insert_these[j]
                        assert ka.is_left
                        s: str = ka.value
                        for k in range(l):
                            ins[j + ni + k] = Sum.inl(f"{k}-{s}")
                    ni += l - 1
                else:
                    if j in self.then_insert_these:
                        ins[j + ni] = self.then_insert_these[j]

            else:
                a: A = ia.value
                ins[j + ni] = ia

        return Reorder(
            l1_len=nr,
            l2_len=ni,
            remove_these=r,
            then_insert_these=ins,
        )

    def __add__(self, other: "Reorder[A]") -> "Reorder[A]":
        return self.concat(other)

    def _map_keys(self, fn: Callable[[str], str]) -> "Reorder[A]":
        return Reorder(
            self.l1_len,
            self.l2_len,
            remove_these={
                i: (fn(s) if s is not None else s) for i, s in self.remove_these.items()
            },
            then_insert_these={
                j: v.mapl(fn) for j, v in self.then_insert_these.items()
            },
        )

    @classmethod
    def identity(cls, length: int) -> "Reorder[Any]":
        return cls(l1_len=length, l2_len=length, remove_these={}, then_insert_these={})

    @property
    def deletions(self):
        """Get the indices of items in the first list that are deleted."""
        for i in self.remove_these.keys():
            if self.remove_these[i] is None:
                yield i

    @property
    def creations(self):
        """Get the indices of items in the second list that are created."""
        for j, v in self.then_insert_these.items():
            if not v.is_left:
                yield j

    @property
    def moves(self) -> Iterable[tuple[int, int]]:
        """Get pairs of i,j indices of all pairs of values that appear in both lists."""
        rm = {}
        count = 0
        for i in range(self.l1_len):
            if i in self.remove_these:
                t = self.remove_these[i]
                if t is None:
                    continue
                else:
                    assert t not in rm
                    rm[t] = i
            else:
                assert count not in rm
                rm[count] = i
                count += 1
        rm_count = count
        count = 0
        for j in range(self.l2_len):
            if j in self.then_insert_these:
                v = self.then_insert_these[j]
                if v.is_left:
                    k = v.value
                    i = rm.pop(k)
                    yield (i, j)
                else:
                    continue
            else:
                i = rm.pop(count)
                yield (i, j)
                count += 1
        assert count == rm_count
        assert len(rm) == 0, f"Remaining: {rm}, {self}"


def deduplicate_values(d: dict):
    acc = {}
    c = Counter()
    for k, v in d.items():
        n = c[v]
        acc[k] = f"{v}-{n}"
        c[v] += 1
    return acc


def diff(l1: list[A], l2: list[A]) -> Reorder[A]:
    removes = dict()
    co_removes = dict()
    codes = SequenceMatcher(None, l1, l2).get_opcodes()
    for tag, i1, i2, j1, j2 in codes:
        if tag == "replace" or tag == "delete":
            for i in range(i1, i2):
                removes[i] = str(hash(l1[i]))
        if tag == "replace" or tag == "insert":
            for j in range(j1, j2):
                assert j not in co_removes
                co_removes[j] = str(hash(l2[j]))
    # deduplicate
    removes = deduplicate_values(removes)
    co_removes = deduplicate_values(co_removes)
    # if a new item is inserted we need to include a copy of the item.
    new_inserts = set(co_removes.values()).difference(removes.values())
    deletions = set(removes.values()).difference(co_removes.values())
    remove_these: Any = {i: (None if h in deletions else h) for i, h in removes.items()}
    inserts = {
        i: (Sum.inr(l2[i]) if h in new_inserts else Sum.inl(h))
        for i, h in co_removes.items()
    }

    return Reorder(
        len(l1), len(l2), remove_these=remove_these, then_insert_these=inserts
    )



from functools import partial, singledispatch
from typing import Any, Literal, Optional, ParamSpec, Sequence, TypeVar, Union, overload

from .textnode import TextNodeSpec
from .vdom import Html, NormSpec, normalise_html
from .element import ElementSpec
from .fiber import Component, FiberSpec


P = ParamSpec("P")


@overload
def h(tag: str, attrs: dict, *children: Html, key: Optional[str] = None) -> ElementSpec:
    ...


@overload
def h(tag: Component[P], *args: P.args, **kwargs: P.kwargs) -> FiberSpec[P]:
    ...


def h(tag, attrs, *children: Html, key=None, **kwargs) -> Union[ElementSpec, FiberSpec]:  # type: ignore
    if type(tag) == str:
        if not isinstance(attrs, dict):
            raise TypeError("attrs must be a dict")
        if len(kwargs) > 0:
            raise ValueError("kwargs are not supported for tags")
        if key is not None:
            attrs["key"] = key
        all_children = normalise_html(list(children))
        return ElementSpec(tag=tag, attrs=attrs, children=all_children)
    elif callable(tag):
        args = [attrs, *children]
        return FiberSpec(component=tag, props_args=args, key=key, props_kwargs=kwargs)
    else:
        raise TypeError(f"unrecognised tag: {tag}")


# [todo] type-safe html elements.
# [todo] type-safe CSS inline styles.

div = partial(h, tag="div")
p = partial(h, tag="p")
h1 = partial(h, tag="h1")
h2 = partial(h, tag="h2")



from dataclasses import dataclass, field
from typing import Any, Union

from .rendering import Rendering

"""
[todo] ABC
[todo] serialisation
[todo] patch application
"""


@dataclass
class InvalidatePatch:
    """Used to trigger a re-render of entire tree. Avoid."""

    kind: str = field(default="invalidate")

    @property
    def is_empty(self) -> bool:
        return False


@dataclass
class ModifyAttributesPatch:
    remove: list[str]
    add: dict[str, Any]
    element_id: str
    kind: str = field(default="modify-attrs")

    @property
    def is_empty(self) -> bool:
        return len(self.remove) == 0 and len(self.add) == 0


@dataclass
class ModifyChildrenPatch:
    element_id: str
    children_length_start: int
    remove_these: dict
    then_insert_these: dict
    kind: str = field(default="modify-children")

    @property
    def is_empty(self) -> bool:
        return len(self.remove_these) == 0 and len(self.then_insert_these) == 0


@dataclass
class ReplaceElementPatch:
    element_id: str
    new_element: Rendering  # output of render
    kind: str = field(default="replace-element")

    @property
    def is_empty(self) -> bool:
        return False


@dataclass
class ReplaceRootPatch:
    items: list[Rendering]  # output of render
    kind: str = field(default="replace-root")

    @property
    def is_empty(self) -> bool:
        return False


# [todo] RPC-encoding for patches
Patch = Union[
    ModifyAttributesPatch,
    ModifyChildrenPatch,
    ReplaceRootPatch,
    ReplaceElementPatch,
    InvalidatePatch,
]



from dataclasses import dataclass
from typing import ClassVar, Union
from .rendering import EventHandler, RenderedAttrVal
from miniscutil import dict_diff
import logging
from .patch import ModifyAttributesPatch, ModifyChildrenPatch, ReplaceElementPatch
from .vdom import (
    Id,
    NormSpec,
    Vdom,
    VdomContext,
    create,
    dispose,
    fresh_id,
    patch,
    reconcile_lists,
    vdom_context,
)
from .rendering import Rendering, RenderedElement

logger = logging.getLogger("reactor")


@dataclass
class ElementSpec(NormSpec):
    tag: str
    attrs: dict
    children: list[NormSpec]

    @property
    def key(self):
        return hash(("element", self.tag, self.attrs.get("key", None)))

    def __str__(self) -> str:
        return f"<{self.tag}/>"

    def create(self) -> "Element":
        return Element.create(self)


@dataclass
class Element(Vdom):
    spec_type: ClassVar = ElementSpec
    tag: str
    attrs: dict[str, RenderedAttrVal]
    children: list[Vdom]
    id: Id
    key: int

    def __str__(self) -> str:
        return f"<{self.tag} {self.id}>"

    @classmethod
    def create(cls, spec: ElementSpec):
        id = fresh_id()
        attrs = {}
        for k, v in spec.attrs.items():
            if callable(v):
                handler_id = f"{id}/{k}"
                vdom_context.get()._register_event(handler_id, v)
                v = EventHandler(handler_id)
            attrs[k] = v
        elt = cls(
            spec.tag, attrs=attrs, children=create(spec.children), id=id, key=spec.key
        )
        return elt

    def dispose(self):
        # delete references to event handlers.
        for k, v in self.attrs.items():
            if isinstance(v, EventHandler):
                vdom_context.get()._unregister_event(v.handler_id)
        dispose(self.children)

    def render(self) -> Rendering:
        return RenderedElement(
            id=self.id,
            tag=self.tag,
            attrs=self.attrs,
            children=[c.render() for c in self.children],
        )

    def reconcile_attrs(self, new_attrs_spec: dict) -> dict:
        for k, v in self.attrs.items():
            if isinstance(v, EventHandler):
                vdom_context.get()._unregister_event(v.handler_id)
        new_attrs = {}
        for k, v in new_attrs_spec.items():
            if callable(v):
                handler_id = f"{self.id}/{k}"
                vdom_context.get()._register_event(handler_id, v)
                v = EventHandler(handler_id)
            new_attrs[k] = v
        diff = dict_diff(self.attrs, new_attrs)
        remove = list(diff.rm)
        add = {k: new_attrs[k] for k in diff.add}
        mod = {k: v2 for k, (v1, v2) in diff.mod.items()}
        patch(
            ModifyAttributesPatch(remove=remove, add={**add, **mod}, element_id=self.id)
        )
        self.attrs = new_attrs
        return self.attrs

    def reconcile(self, new_spec: ElementSpec) -> "Element":
        assert isinstance(new_spec, ElementSpec)
        logger.debug(f"reconcile {str(self)} ← {str(new_spec)}")
        if (self.key != new_spec.key) or (self.tag != new_spec.tag):
            self.dispose()
            v = new_spec.create()
            logger.debug(f"replacing {str(self)} → {str(v)}")
            patch(ReplaceElementPatch(self.id, v.render()))
            return v
        self.reconcile_attrs(new_spec.attrs)
        children, r = reconcile_lists(self.children, new_spec.children)
        # [todo] apply expander to r here.
        patch(
            ModifyChildrenPatch(
                children_length_start=r.l1_len,
                element_id=self.id,
                remove_these=r.remove_these,
                then_insert_these=r.then_insert_these,
            )
        )
        self.tag = new_spec.tag
        self.children = children
        return self



import asyncio
from dataclasses import dataclass
import inspect
import logging
from typing import Any, Callable, Optional
from .html import Html
from .vdom import (
    Id,
    Rendering,
    Vdom,
    dispose,
    fresh_id,
    reconcile_lists,
    create,
    render,
    normalise_html,
    set_vdom_context,
)
from .patch import ModifyChildrenPatch, Patch
from miniscutil.asyncio_helpers import MessageQueue

logger = logging.getLogger("uxu")


@dataclass
class EventArgs:
    handler_id: str
    name: str
    params: Optional[Any]


class Manager:
    id: Id

    event_table: dict[str, Callable]
    event_tasks: set[asyncio.Task]
    root: list[Vdom]
    pending_patches: MessageQueue[Patch]

    def __init__(self):
        self.id = fresh_id()
        self.event_table = {}
        self.pending_patches = MessageQueue()

    def initialize(self, html: Html):
        with set_vdom_context(self):
            spec = normalise_html(html)
            self.root = create(spec)

    def update(self, html: Html):
        if not self.is_initialized:
            return self.initialize(html)
        with set_vdom_context(self):
            spec = normalise_html(html)
            new_root, reorder = reconcile_lists(self.root, spec)
            self.root = new_root
            self._patch(
                ModifyChildrenPatch(
                    element_id=self.id,
                    remove_these=reorder.remove_these,
                    then_insert_these=reorder.then_insert_these,
                    children_length_start=reorder.l1_len,
                )
            )

    @property
    def is_initialized(self):
        return hasattr(self, "root")

    def render(self) -> list[Rendering]:
        self.pending_patches.clear()
        return render(self.root)

    def dispose(self):
        dispose(self.root)
        delattr(self, "root")
        self.pending_patches.clear()

    def _patch(self, patch: Patch):
        if patch.is_empty:
            return
        self.pending_patches.push(patch)

    def _register_event(self, k: str, handler: Callable):
        assert k not in self.event_table
        self.event_table[k] = handler

    def _unregister_event(self, k: str):
        assert k in self.event_table
        del self.event_table[k]

    def handle_event(self, params: EventArgs):
        with set_vdom_context(self):
            assert isinstance(params, EventArgs)
            logger.debug(f"handling {params.handler_id}")
            k = params.handler_id
            if k not in self.event_table:
                logger.debug(f"No handler for {params.handler_id}")
                return
            assert k in self.event_table
            handler = self.event_table[k]

            r = handler(params.params)
            if inspect.iscoroutine(r):
                et = asyncio.create_task(r)
                self.event_tasks.add(et)
                et.add_done_callback(self.event_tasks.discard)
                # note that we don't cancel event tasks if the handler gets replaced.

        # event handler will call code to invalidate components.
        # [todo] trigger a re-render

    async def wait_patches(self):
        return await self.pending_patches.pop_many()
